""" This module provides the infrastructure for loopy preambles that have
a complex requirement structure. This includes:
* preambles triggering the creation of other preambles
* a caching mechanism to avoid duplicated preambles where harmful
"""

from __future__ import absolute_import
from pytools import memoize


# have one cache the module level. It is easier than handing around an instance of it.
_cache = {}


def _freeze(data):
    """ A function that deterministically generates an
    immutable item from the given data. Used for cache key generation.
    """
    from collections import Hashable, Iterable, MutableMapping

    # Strings are iterable and hashable, but they do not pose any problems.
    if isinstance(data, str):
        return data

    import ufl.classes
    if isinstance(data, ufl.classes.Expr):
        return data

    from pymbolic.primitives import Expression
    if isinstance(data, Expression):
        return data

    # Check if the given data is already hashable
    if isinstance(data, Hashable):
        if isinstance(data, Iterable):
            return type(data)(_freeze(i) for i in data)
        return data

    # convert standard mutable containers
    if isinstance(data, MutableMapping):
        return tuple((_freeze(k), _freeze(v)) for k, v in data.iteritems())
    if isinstance(data, Iterable):
        return tuple(_freeze(i) for i in data)

    # we don't know how to handle this object, so we give up
    raise TypeError('Cannot freeze non-hashable object {} of type {}'.format(data, type(data)))


class _GlobalCounter(object):
    counter = 0

    def get(self):
        _GlobalCounter.counter = _GlobalCounter.counter + 1
        return _GlobalCounter.counter


def no_caching(*a, **k):
    return _GlobalCounter().get()


class _CacheItemMeta(type):
    """ A meta class for cache items. Keyword arguments are forwarded
    th decorator factory below (check the documentation there)
    """
    def __new__(cls, name, bases, d, counted=False, on_store=lambda x: x, item_tags=[]):
        rettype = type(name, bases, d)
        if counted:
            original_on_store = on_store

            def add_count(x):
                count = _GlobalCounter().get()
                if isinstance(x, tuple):
                    return (count, original_on_store(*x))
                else:
                    return (count, original_on_store(x))
            on_store = add_count

        def _init(s, x):
            if isinstance(x, tuple) and not counted:
                s.content = on_store(*x)
            else:
                s.content = on_store(x)
            s.tags = item_tags
            s.counted = counted

        setattr(rettype, '__init__', _init)

        return rettype


@memoize(use_kwargs=True)
def _construct_cache_item_type(name, **kwargs):
    """ Wrap the generation of cache item types from the meta class.
    At the same time, memoization assures that types are the same for
    multiple cache items
    """
    return _CacheItemMeta.__new__(_CacheItemMeta, name, (), {}, **kwargs)


class _RegisteredFunction(object):
    """ The data structure for a function that accesses UFL2LoopyDataCache """
    def __init__(self, func,
                 cache_key_generator=lambda *a, **kw: a,
                 **kwargs
                 ):
        self.func = func
        self.cache_key_generator = cache_key_generator

        # Append the current context item tags to the given ones
        from dune.perftool.generation.context import get_context_tags
        if 'item_tags' in kwargs:
            kwargs['item_tags'] = tuple(kwargs['item_tags']) + get_context_tags()
        else:
            kwargs['item_tags'] = get_context_tags()

        self.itemtype = _construct_cache_item_type("CacheItemType", **kwargs)

    def __call__(self, *args, **kwargs):
        # Get the cache key from the given arguments
        cache_args = self.cache_key_generator(*args, **kwargs)
        # Make sure that all keyword arguments have vanished from the cache_args
        assert (lambda *a, **k: len(k) == 0)(cache_args)
        cache_key = (self, _freeze(self.cache_key_generator(*args, **kwargs)))
        # check whether we have a cache hit
        if cache_key in _cache:
            # and return the result depending on the cache item type
            return _cache[cache_key].content
        else:
            # evaluate the original function and wrap it in a cache item
            citem = self.itemtype(self.func(*args, **kwargs))
            _cache[cache_key] = citem
            return citem.content


def generator_factory(**factory_kwargs):
    """ A function decorator factory

    Generates a function decorator, that turns a given function into
    a function that stores its result in a data cache.

    The generated decorator may be used with or without keyword arguments.

    The keyword arguments given to this factory override any keywords
    given to the decorator. This allows the definition of specialized
    decorators for different purposes.

    Possible keyword arguments:
    ---------------------------
    cache_key_generator : function
        A function that maps the arguments of the function to a subset that
        determines whether to use a caches result. The return type is arbitrary,
        as it will be turned immutable by the cache machine afterwards.
        Defaults to identity.
    item_tags : tuple
        A tuple of tags (simple strings) to give to the cache items. Items can be
        retrieved and deleted by tag.
    on_store : function
        A function to apply to the return value of the decorated function
        before storing in the cache. May be used to apply wrappers.
    counted : bool
        Will add a counted tag to the cache item. The type of the stored
        items automatically turns to a tuple with the counttag being the first entry.
    no_deco : bool
        Instead of a decorator, return a function that uses identity as a body.
    """
    # Tuplize the item_tags parameter
    if "item_tags" in factory_kwargs and isinstance(factory_kwargs["item_tags"], str):
        factory_kwargs["item_tags"] = (factory_kwargs["item_tags"],)

    no_deco = factory_kwargs.pop("no_deco", False)

    def _dec(*args, **kwargs):
        # Modify the kwargs according to the factorys kwargs
        for k in factory_kwargs:
            kwargs[k] = factory_kwargs[k]
        if args:
            assert len(args) == 1
            return _RegisteredFunction(args[0], **kwargs)
        else:
            return lambda f: _RegisteredFunction(f, **kwargs)

    if no_deco:
        return _dec(lambda x: x)
    else:
        return _dec


# define a decorator 'cached', which may be used as a standard decorator, if no further
# magic is to be added to the caching mechanism
cached = generator_factory(item_tags="default_cached")


class _ConditionDict(dict):
        def __init__(self, tags):
            dict.__init__(self)
            self.tags = tags

        def __getitem__(self, i):
            # If we do not add these special cases the dictionary will return False
            # when we execute the following code:
            #
            # eval ("True", _ConditionDict(v.tags)
            #
            # But in this case we want to return True! A normal dictionary would not attempt
            # to replace "True" if "True" is not a key. The _ConditionDict obviously has no
            # such concerns ;).
            if i == "True":
                return True
            if i == "False":
                return False
            return i in self.tags


def _filter_cache_items(condition):
    return {k: v for k, v in _cache.items() if eval(condition, _ConditionDict(v.tags))}


def retrieve_cache_items(condition=True, make_generable=False):
    choice = _filter_cache_items(condition).values()

    def as_generable(content):
        if make_generable:
            from cgen import Generable, Line
            if isinstance(content, Generable):
                return content
            if isinstance(content, str):
                return Line(text=content + '\n')
            assert False
        else:
            return content

    # First yield all those items that are not sorted
    for item in choice:
        if not item.counted:
            yield as_generable(item.content)

    # And now the sorted ones
    for item in sorted([i for i in choice if i.counted], key=lambda i: i.content[0]):
        from collections import Iterable
        if isinstance(item.content[1], Iterable) and not isinstance(item.content[1], str):
            for l in item.content[1]:
                yield as_generable(l)
        else:
            yield as_generable(item.content[1])


def delete_cache_items(condition=True, keep=False):
    """ Delete items from the cache. """
    if not keep:
        condition = "not ({})".format(condition)

    global _cache
    _cache = _filter_cache_items(condition)
